/*-------------------------------------------------------------------------
 *
 * walsender.c
 *
 * The WAL sender process (walsender) is new as of Postgres 9.0. It takes
 * care of sending XLOG from the primary server to a single recipient.
 * (Note that there can be more than one walsender process concurrently.)
 * It is started by the postmaster when the walreceiver of a standby server
 * connects to the primary server and requests XLOG streaming replication.
 * It attempts to keep reading XLOG records from the disk and sending them
 * to the standby server, as long as the connection is alive (i.e., like
 * any backend, there is a one-to-one relationship between a connection
 * and a walsender process).
 *
 * Normal termination is by SIGTERM, which instructs the walsender to
 * close the connection and exit(0) at next convenient moment. Emergency
 * termination is by SIGQUIT; like any backend, the walsender will simply
 * abort and exit on SIGQUIT. A close of the connection and a FATAL error
 * are treated as not a crash but approximately normal termination;
 * the walsender will exit quickly without sending any more XLOG records.
 *
 * If the server is shut down, postmaster sends us SIGUSR2 after all
 * regular backends have exited and the shutdown checkpoint has been written.
 * This instruct walsender to send any outstanding WAL, including the
 * shutdown checkpoint record, and then exit.
 *
 *
 * Portions Copyright (c) 2010-2010, PostgreSQL Global Development Group
 *
 * IDENTIFICATION
 *	  $PostgreSQL: pgsql/src/backend/replication/walsender.c,v 1.32 2010/09/15 06:51:19 heikki Exp $
 *
 *-------------------------------------------------------------------------
 */
#include "postgres.h"

#include <signal.h>
#include <unistd.h>

#include "access/xlog_internal.h"
#include "catalog/pg_type.h"
#include "libpq/libpq.h"
#include "libpq/pqformat.h"
#include "libpq/pqsignal.h"
#include "miscadmin.h"
#include "replication/walprotocol.h"
#include "replication/walsender.h"
#include "storage/fd.h"
#include "storage/ipc.h"
#include "storage/pmsignal.h"
#include "storage/proc.h"
#include "tcop/tcopprot.h"
#include "utils/guc.h"
#include "utils/memutils.h"
#include "utils/ps_status.h"


/* Array of WalSnds in shared memory */
WalSndCtlData *WalSndCtl = NULL;

/* My slot in the shared memory array */
static WalSnd *MyWalSnd = NULL;

/* Array of WalSndWaiter in shared memory */
static WalSndWaiter  *WalSndWaiters;

/* Global state */
bool		am_walsender = false;		/* Am I a walsender process ? */

/* User-settable parameters for walsender */
int			max_wal_senders = 0;	/* the maximum number of concurrent walsenders */
int			WalSndDelay = 200;	/* max sleep time between some actions */
int			replication_timeout = 0;	/* maximum time to wait for the Ack from the standby */
char	   *standby_fencing_command = NULL;	/* command to shoot the standby in the head */

/*
 * Buffer for WAL sending
 *
 * WalSndOutBuffer is a work area in which the output message is constructed.
 * It's used in just so we can avoid re-palloc'ing the buffer on each cycle.
 * It must be of size 6 + sizeof(WalDataMessageHeader) + MAX_SEND_SIZE.
 */
static char	   *WalSndOutBuffer;
static int		WalSndOutHead;		/* head of pending output */
static int		WalSndOutTail;		/* tail of pending output */

/*
 * These variables are used similarly to openLogFile/Id/Seg/Off,
 * but for walsender to read the XLOG.
 */
static int	sendFile = -1;
static uint32 sendId = 0;
static uint32 sendSeg = 0;
static uint32 sendOff = 0;

/*
 * How far have we sent WAL already? This is also advertised in
 * MyWalSnd->sentPtr.  (Actually, this is the next WAL location to send.)
 */
static XLogRecPtr sentPtr = {0, 0};

/*
 * How far have we completed replication already? This is also
 * advertised in MyWalSnd->ackdPtr. This is not used in asynchronous
 * replication case.
 */
static XLogRecPtr ackdPtr = {0, 0};

/* Flags set by signal handlers for later service in main loop */
static volatile sig_atomic_t got_SIGHUP = false;
static volatile sig_atomic_t standalone_disabled = false;
static volatile sig_atomic_t shutdown_requested = false;
static volatile sig_atomic_t ready_to_stop = false;

/* Flag set by signal handler of backends for replication */
static volatile sig_atomic_t replication_complete = false;
static volatile sig_atomic_t exit_before_complete = false;

/* Signal handlers */
static void WalSndSigHupHandler(SIGNAL_ARGS);
static void WalSndDisableStandaloneHandler(SIGNAL_ARGS);
static void WalSndShutdownHandler(SIGNAL_ARGS);
static void WalSndQuickDieHandler(SIGNAL_ARGS);
static void WalSndXLogSendHandler(SIGNAL_ARGS);
static void WalSndLastCycleHandler(SIGNAL_ARGS);

/* Prototypes for private functions */
static int	WalSndLoop(void);
static void InitWalSnd(void);
static void WalSndHandshake(void);
static void WalSndKill(int code, Datum arg);
static void XLogRead(char *buf, XLogRecPtr recptr, Size nbytes);
static bool XLogSend(bool *caughtup, bool *pending);
static void ProcessStreamMsgs(StringInfo inMsg);
static void ExecuteStandbyFencingCommand(void);

static void RegisterWalSndWaiter(BackendId backendId, XLogRecPtr record,
								 Latch *latch);
static void WakeupWalSndWaiters(XLogRecPtr record, bool abort);
static XLogRecPtr GetOldestAckdPtr(void);


/* Main entry point for walsender process */
int
WalSenderMain(void)
{
	MemoryContext walsnd_context;

	if (RecoveryInProgress())
		ereport(FATAL,
				(errcode(ERRCODE_CANNOT_CONNECT_NOW),
				 errmsg("recovery is still in progress, can't accept WAL streaming connections")));

	/* Create a per-walsender data structure in shared memory */
	InitWalSnd();

	/*
	 * Create a memory context that we will do all our work in.  We do this so
	 * that we can reset the context during error recovery and thereby avoid
	 * possible memory leaks.  Formerly this code just ran in
	 * TopMemoryContext, but resetting that would be a really bad idea.
	 *
	 * XXX: we don't actually attempt error recovery in walsender, we just
	 * close the connection and exit.
	 */
	walsnd_context = AllocSetContextCreate(TopMemoryContext,
										   "Wal Sender",
										   ALLOCSET_DEFAULT_MINSIZE,
										   ALLOCSET_DEFAULT_INITSIZE,
										   ALLOCSET_DEFAULT_MAXSIZE);
	MemoryContextSwitchTo(walsnd_context);

	/* Unblock signals (they were blocked when the postmaster forked us) */
	PG_SETMASK(&UnBlockSig);

	/* Tell the standby that walsender is ready for receiving commands */
	ReadyForQuery(DestRemote);

	/* Handle handshake messages before streaming */
	WalSndHandshake();

	/* Initialize shared memory status */
	{
		/* use volatile pointer to prevent code rearrangement */
		volatile WalSnd *walsnd = MyWalSnd;

		SpinLockAcquire(&walsnd->mutex);
		walsnd->sentPtr = sentPtr;
		SpinLockRelease(&walsnd->mutex);
	}

	/* Main loop of walsender */
	return WalSndLoop();
}

/*
 * Execute commands from walreceiver, until we enter streaming mode.
 */
static void
WalSndHandshake(void)
{
	StringInfoData input_message;
	bool		replication_started = false;

	initStringInfo(&input_message);

	while (!replication_started)
	{
		int			firstchar;

		/* Wait for a command to arrive */
		firstchar = pq_getbyte();

		/*
		 * Emergency bailout if postmaster has died.  This is to avoid the
		 * necessity for manual cleanup of all postmaster children.
		 */
		if (!PostmasterIsAlive(true))
			exit(1);

		/*
		 * Check for any other interesting events that happened while we
		 * slept.
		 */
		if (got_SIGHUP)
		{
			got_SIGHUP = false;
			ProcessConfigFile(PGC_SIGHUP);
		}

		if (firstchar != EOF)
		{
			/*
			 * Read the message contents. This is expected to be done without
			 * blocking because we've been able to get message type code.
			 */
			if (pq_getmessage(&input_message, 0))
				firstchar = EOF;	/* suitable message already logged */
		}

		/* Handle the very limited subset of commands expected in this phase */
		switch (firstchar)
		{
			char		modestr[6];

			case 'Q':			/* Query message */
				{
					const char *query_string;
					XLogRecPtr	recptr;

					query_string = pq_getmsgstring(&input_message);
					pq_getmsgend(&input_message);

					if (strcmp(query_string, "IDENTIFY_SYSTEM") == 0)
					{
						StringInfoData buf;
						char		sysid[32];
						char		tli[11];

						/*
						 * Reply with a result set with one row, two columns.
						 * First col is system ID, and second is timeline ID
						 */

						snprintf(sysid, sizeof(sysid), UINT64_FORMAT,
								 GetSystemIdentifier());
						snprintf(tli, sizeof(tli), "%u", ThisTimeLineID);

						/* Send a RowDescription message */
						pq_beginmessage(&buf, 'T');
						pq_sendint(&buf, 2, 2); /* 2 fields */

						/* first field */
						pq_sendstring(&buf, "systemid");		/* col name */
						pq_sendint(&buf, 0, 4); /* table oid */
						pq_sendint(&buf, 0, 2); /* attnum */
						pq_sendint(&buf, TEXTOID, 4);	/* type oid */
						pq_sendint(&buf, -1, 2);		/* typlen */
						pq_sendint(&buf, 0, 4); /* typmod */
						pq_sendint(&buf, 0, 2); /* format code */

						/* second field */
						pq_sendstring(&buf, "timeline");		/* col name */
						pq_sendint(&buf, 0, 4); /* table oid */
						pq_sendint(&buf, 0, 2); /* attnum */
						pq_sendint(&buf, INT4OID, 4);	/* type oid */
						pq_sendint(&buf, 4, 2); /* typlen */
						pq_sendint(&buf, 0, 4); /* typmod */
						pq_sendint(&buf, 0, 2); /* format code */
						pq_endmessage(&buf);

						/* Send a DataRow message */
						pq_beginmessage(&buf, 'D');
						pq_sendint(&buf, 2, 2); /* # of columns */
						pq_sendint(&buf, strlen(sysid), 4);		/* col1 len */
						pq_sendbytes(&buf, (char *) &sysid, strlen(sysid));
						pq_sendint(&buf, strlen(tli), 4);		/* col2 len */
						pq_sendbytes(&buf, (char *) tli, strlen(tli));
						pq_endmessage(&buf);

						/* Send CommandComplete and ReadyForQuery messages */
						EndCommand("SELECT", DestRemote);
						ReadyForQuery(DestRemote);
						/* ReadyForQuery did pq_flush for us */
					}
					else if (sscanf(query_string, "START_REPLICATION %X/%X MODE %5s",
									&recptr.xlogid, &recptr.xrecoff, modestr) == 3)
					{
						StringInfoData buf;
						int			mode;

						/*
						 * Check that we're logging enough information in the
						 * WAL for log-shipping.
						 *
						 * NOTE: This only checks the current value of
						 * wal_level. Even if the current setting is not
						 * 'minimal', there can be old WAL in the pg_xlog
						 * directory that was created with 'minimal'. So this
						 * is not bulletproof, the purpose is just to give a
						 * user-friendly error message that hints how to
						 * configure the system correctly.
						 */
						if (wal_level == WAL_LEVEL_MINIMAL)
							ereport(FATAL,
									(errcode(ERRCODE_CANNOT_CONNECT_NOW),
									 errmsg("standby connections not allowed because wal_level=minimal")));

						/* Verify that the specified replication mode is valid */
						{
							const struct config_enum_entry *entry;

							for (entry = replication_mode_options; entry && entry->name; entry++)
							{
								if (strcmp(modestr, entry->name) == 0)
								{
									mode = entry->val;
									break;
								}
							}
							if (entry == NULL || entry->name == NULL)
								ereport(FATAL,
										(errcode(ERRCODE_PROTOCOL_VIOLATION),
										 errmsg("invalid replication mode: %s", modestr)));
						}

						/* Change the state according to replication mode specified by standby */
						{
							/* use volatile pointer to prevent code rearrangement */
							volatile WalSnd *walsnd = MyWalSnd;

							SpinLockAcquire(&walsnd->mutex);
							walsnd->walSndState = (mode == REPLICATION_MODE_ASYNC) ?
								WALSND_ASYNC : WALSND_CATCHUP;
							SpinLockRelease(&walsnd->mutex);
						}

						/* Send a CopyBothResponse message, and start streaming */
						pq_beginmessage(&buf, 'W');
						pq_sendbyte(&buf, 0);
						pq_sendint(&buf, 0, 2);
						pq_endmessage(&buf);
						pq_flush();

						/*
						 * Initialize positions to the received one, then the
						 * xlog records begin to be shipped from that position
						 */
						sentPtr = ackdPtr = recptr;

						/* break out of the loop */
						replication_started = true;
					}
					else
					{
						ereport(FATAL,
								(errcode(ERRCODE_PROTOCOL_VIOLATION),
								 errmsg("invalid standby query string: %s", query_string)));
					}
					break;
				}

			case 'X':
				/* standby is closing the connection */
				proc_exit(0);

			case EOF:
				/* standby disconnected unexpectedly */
				ereport(COMMERROR,
						(errcode(ERRCODE_PROTOCOL_VIOLATION),
						 errmsg("unexpected EOF on standby connection")));
				proc_exit(0);

			default:
				ereport(FATAL,
						(errcode(ERRCODE_PROTOCOL_VIOLATION),
						 errmsg("invalid standby handshake message type %d", firstchar)));
		}
	}
}

/*
 * Process messages received from the standby.
 *
 * ereports on error.
 */
static void
ProcessStreamMsgs(StringInfo inMsg)
{
	bool	acked = false;

	/* Loop to process successive complete messages available */
	for (;;)
	{
		unsigned char firstchar;
		int			r;

		r = pq_getbyte_if_available(&firstchar);
		if (r < 0)
		{
			/* unexpected error or EOF */
			ereport(COMMERROR,
					(errcode(ERRCODE_PROTOCOL_VIOLATION),
					 errmsg("unexpected EOF on standby connection")));
			proc_exit(0);
		}
		if (r == 0)
		{
			/* no data available without blocking */
			break;
		}

		/* Handle the very limited subset of commands expected in this phase */
		switch (firstchar)
		{
			case 'd':       /* CopyData message */
			{
				unsigned char	rpltype;

				/*
				 * Read the message contents. This is expected to be done without
				 * blocking because we've been able to get message type code.
				 */
				if (pq_getmessage(inMsg, 0))
					proc_exit(0);		/* suitable message already logged */

				/* Read the replication message type from CopyData message */
				rpltype = pq_getmsgbyte(inMsg);
				switch (rpltype)
				{
					case 'l':
					{
						WalAckMessageData  *msgdata;

						msgdata = (WalAckMessageData *) pq_getmsgbytes(inMsg, sizeof(WalAckMessageData));

						/*
						 * Update local status.
						 *
						 * The ackd ptr received from standby should not
						 * go backwards.
						 */
						if (XLByteLE(ackdPtr, msgdata->ackEnd))
							ackdPtr = msgdata->ackEnd;
						else
							ereport(FATAL,
									(errmsg("replication completion location went back from "
											"%X/%X to %X/%X",
											ackdPtr.xlogid, ackdPtr.xrecoff,
											msgdata->ackEnd.xlogid, msgdata->ackEnd.xrecoff)));

						acked = true;	/* also need to update shared position */
						break;
					}
					default:
						ereport(FATAL,
								(errcode(ERRCODE_PROTOCOL_VIOLATION),
								 errmsg("invalid replication message type %d",
										rpltype)));
				}
				break;
			}

			/*
			 * 'X' means that the standby is closing down the socket.
			 */
			case 'X':
				proc_exit(0);

			default:
				ereport(FATAL,
						(errcode(ERRCODE_PROTOCOL_VIOLATION),
						 errmsg("invalid standby message type %d",
								firstchar)));
		}
	}

	if (acked)
	{
		/* use volatile pointer to prevent code rearrangement */
		volatile WalSnd *walsnd = MyWalSnd;

		SpinLockAcquire(&walsnd->mutex);
		walsnd->ackdPtr = ackdPtr;
		SpinLockRelease(&walsnd->mutex);
	}

	/* Wake up the backends that this walsender had been blocking */
	if (MyWalSnd->walSndState >= WALSND_CATCHUP)
		WakeupWalSndWaiters(ackdPtr, false);
}

/* Main loop of walsender process */
static int
WalSndLoop(void)
{
	StringInfoData	input_message;
	bool		caughtup = false;
	bool		pending = false;
	XLogRecPtr	switchptr = {0, 0};

	initStringInfo(&input_message);

	/*
	 * Allocate buffer that will be used for each output message.  We do this
	 * just once to reduce palloc overhead.  The buffer must be made large
	 * enough for maximum-sized messages.
	 */
	WalSndOutBuffer = palloc(6 + sizeof(WalDataMessageHeader) + MAX_SEND_SIZE);
	WalSndOutHead = WalSndOutTail = 0;

	/* Loop forever, unless we get an error */
	for (;;)
	{
		/*
		 * Emergency bailout if postmaster has died.  This is to avoid the
		 * necessity for manual cleanup of all postmaster children.
		 */
		if (!PostmasterIsAlive(true))
			exit(1);

		/* Process any requests or signals received recently */
		if (got_SIGHUP)
		{
			got_SIGHUP = false;
			ProcessConfigFile(PGC_SIGHUP);
		}

		/*
		 * When SIGUSR2 arrives, we send all outstanding logs up to the
		 * shutdown checkpoint record (i.e., the latest record) and exit.
		 */
		if (ready_to_stop)
		{
			if (!XLogSend(&caughtup, &pending))
				break;
			if (caughtup && !pending)
				shutdown_requested = true;
		}

		/* Normal exit from the walsender is here */
		if (shutdown_requested)
		{
			/* Inform the standby that XLOG streaming was done */
			pq_puttextmessage('C', "COPY 0");
			pq_flush();

			proc_exit(0);
		}

		/*
		 * If we had sent all accumulated WAL in last round or could not
		 * flush pending WAL in output buffer because the socket was not
		 * writable, nap for the configured time before retrying.
		 */
		if (caughtup || pending)
		{
			/*
			 * Even if we wrote all the WAL that was available when we started
			 * sending, more might have arrived while we were sending this
			 * batch. We had the latch set while sending, so we have not
			 * received any signals from that time. Let's arm the latch
			 * again, and after that check that we're still up-to-date.
			 */
			ResetLatch(&MyWalSnd->latch);

			if (!XLogSend(&caughtup, &pending))
				break;

			/*
			 * If the standby has almost caught up with the primary, we change
			 * the state to WALSND_PRESYNC and start making transactions wait
			 * until their WAL has been replicated.
			 *
			 * No lock is required to get WalSnd->walSndState here since it can
			 * be updated only by walsender.
			 */
			if (MyWalSnd->walSndState == WALSND_CATCHUP && caughtup)
			{
				/* use volatile pointer to prevent code rearrangement */
				volatile WalSnd *walsnd = MyWalSnd;

				SpinLockAcquire(&walsnd->mutex);
				walsnd->walSndState = WALSND_PRESYNC;
				SpinLockRelease(&walsnd->mutex);

				/*
				 * switchptr indicates how far we must complete replication
				 * before advertising that the standby has already been in
				 * sync with the primary.
				 */
				switchptr = GetFlushRecPtr();
			}

			if ((caughtup || pending) && !got_SIGHUP && !ready_to_stop &&
				!shutdown_requested)
			{
				bool		check_timeout;
				long		sleeptime;
				int			res;

				/*
				 * XXX: We don't really need the periodic wakeups anymore,
				 * WaitLatchOrSocket should reliably wake up as soon as
				 * something interesting happens.
				 */

				/*
				 * Check for replication timeout if it's enabled and we need
				 * to wait until the socket has become writable to flush
				 * pending WAL in output buffer or until the Ack message
				 * from the standby has become available.
				 */
				if (replication_timeout > 0 &&
					(pending ||
					 (MyWalSnd->walSndState >= WALSND_CATCHUP &&
					  XLByteLT(ackdPtr, sentPtr))))
				{
					sleeptime = replication_timeout;
					check_timeout = true;
				}
				else
				{
					sleeptime = WalSndDelay;
					check_timeout = false;
				}

				/* Sleep */
				res = WaitLatchOrSocket(&MyWalSnd->latch, MyProcPort->sock,
										true, (WalSndOutTail > 0),
										sleeptime * 1000L);

				if (res == 0 && check_timeout)
				{
					/*
					 * Since typically expiration of replication timeout means
					 * communication problem, we don't send the error message
					 * to the standby.
					 */
					ereport(COMMERROR,
							(errmsg("terminating walsender process due to replication timeout")));
					break;
				}
			}

			/* Process messages received from the standby */
			ProcessStreamMsgs(&input_message);

			/*
			 * If the standby has caught up with the primary, we change
			 * the state to WALSND_SYNC and inform the standby that it's
			 * in sync with the primary. This state ensures that all the
			 * transactions completed from a client's point of view have
			 * been replicated to the standby.
			 */
			if (MyWalSnd->walSndState == WALSND_PRESYNC &&
				XLByteLE(switchptr, ackdPtr) && !pending)
			{
				/* use volatile pointer to prevent code rearrangement */
				volatile WalSnd *walsnd = MyWalSnd;

				SpinLockAcquire(&walsnd->mutex);
				walsnd->walSndState = WALSND_SYNC;
				SpinLockRelease(&walsnd->mutex);

				/*
				 * We can send a XLogCatchupComplete message without blocking
				 * since it's guaranteed that there is no pending data in the
				 * output buffer.
				 */
				pq_putmessage('d', "c", 1);
				if (pq_flush())
					break;
			}
		}
		else
		{
			/* Attempt to send the log once every loop */
			if (!XLogSend(&caughtup, &pending))
				break;
		}
	}

	/*
	 * Get here on send failure.  Clean up and exit.
	 *
	 * Reset whereToSendOutput to prevent ereport from attempting to send any
	 * more messages to the standby.
	 */
	if (whereToSendOutput == DestRemote)
		whereToSendOutput = DestNone;

	proc_exit(0);
	return 1;					/* keep the compiler quiet */
}

/* Initialize a per-walsender data structure for this walsender process */
static void
InitWalSnd(void)
{
	int			i;

	/*
	 * WalSndCtl should be set up already (we inherit this by fork() or
	 * EXEC_BACKEND mechanism from the postmaster).
	 */
	Assert(WalSndCtl != NULL);
	Assert(MyWalSnd == NULL);

	/*
	 * Find a free walsender slot and reserve it. If this fails, we must be
	 * out of WalSnd structures.
	 */
	for (i = 0; i < max_wal_senders; i++)
	{
		/* use volatile pointer to prevent code rearrangement */
		volatile WalSnd *walsnd = &WalSndCtl->walsnds[i];

		SpinLockAcquire(&walsnd->mutex);

		if (walsnd->pid != 0)
		{
			SpinLockRelease(&walsnd->mutex);
			continue;
		}
		else
		{
			/*
			 * Found a free slot. Reserve it for us.
			 */
			walsnd->pid = MyProcPid;
			MemSet(&walsnd->sentPtr, 0, sizeof(XLogRecPtr));
			MemSet(&walsnd->ackdPtr, 0, sizeof(XLogRecPtr));
			walsnd->walSndState = WALSND_INIT;
			SpinLockRelease(&walsnd->mutex);
			/* don't need the lock anymore */
			OwnLatch((Latch *) &walsnd->latch);
			MyWalSnd = (WalSnd *) walsnd;

			break;
		}
	}
	if (MyWalSnd == NULL)
		ereport(FATAL,
				(errcode(ERRCODE_TOO_MANY_CONNECTIONS),
				 errmsg("number of requested standby connections "
						"exceeds max_wal_senders (currently %d)",
						max_wal_senders)));

	/* Arrange to clean up at walsender exit */
	on_shmem_exit(WalSndKill, 0);
}

/* Destroy the per-walsender data structure for this walsender process */
static void
WalSndKill(int code, Datum arg)
{
	/* use volatile pointer to prevent code rearrangement */
	volatile WalSnd *walsnd = MyWalSnd;
	XLogRecPtr		oldestPtr;
	WalSndState		saveState;

	Assert(MyWalSnd != NULL);

	/*
	 * If replication was terminated for a reason other than the master
	 * server shutdown or emergency bailout (i.e., unexpected death of
	 * postmaster), we can expect this server can work standalone,
	 * so we call standby_fencing_command to shoot the standby server
	 * in the head if it's specified.
	 */
	if (!standalone_disabled && PostmasterIsAlive(true))
		ExecuteStandbyFencingCommand();

	/*
	 * If this walsender had been the last synchronous one, we wake up
	 * all the backends waiting for replication.
	 */
	saveState = walsnd->walSndState;
	SpinLockAcquire(&walsnd->mutex);
	walsnd->walSndState = WALSND_INIT;
	SpinLockRelease(&walsnd->mutex);
	oldestPtr = GetOldestAckdPtr();
	if (saveState >= WALSND_CATCHUP &&
		oldestPtr.xlogid == 0 && oldestPtr.xrecoff == 0)
		WakeupWalSndWaiters(oldestPtr, standalone_disabled);

	/*
	 * Mark WalSnd struct no longer in use. Assume that no lock is required
	 * for this.
	 */
	MyWalSnd->pid = 0;
	DisownLatch(&MyWalSnd->latch);

	/* WalSnd struct isn't mine anymore */
	MyWalSnd = NULL;
}

/*
 * Read 'nbytes' bytes from WAL into 'buf', starting at location 'recptr'
 *
 * XXX probably this should be improved to suck data directly from the
 * WAL buffers when possible.
 */
static void
XLogRead(char *buf, XLogRecPtr recptr, Size nbytes)
{
	XLogRecPtr	startRecPtr = recptr;
	char		path[MAXPGPATH];
	uint32		lastRemovedLog;
	uint32		lastRemovedSeg;
	uint32		log;
	uint32		seg;

	while (nbytes > 0)
	{
		uint32		startoff;
		int			segbytes;
		int			readbytes;

		startoff = recptr.xrecoff % XLogSegSize;

		if (sendFile < 0 || !XLByteInSeg(recptr, sendId, sendSeg))
		{
			/* Switch to another logfile segment */
			if (sendFile >= 0)
				close(sendFile);

			XLByteToSeg(recptr, sendId, sendSeg);
			XLogFilePath(path, ThisTimeLineID, sendId, sendSeg);

			sendFile = BasicOpenFile(path, O_RDONLY | PG_BINARY, 0);
			if (sendFile < 0)
			{
				/*
				 * If the file is not found, assume it's because the standby
				 * asked for a too old WAL segment that has already been
				 * removed or recycled.
				 */
				if (errno == ENOENT)
				{
					char		filename[MAXFNAMELEN];

					XLogFileName(filename, ThisTimeLineID, sendId, sendSeg);
					ereport(ERROR,
							(errcode_for_file_access(),
							 errmsg("requested WAL segment %s has already been removed",
									filename)));
				}
				else
					ereport(ERROR,
							(errcode_for_file_access(),
							 errmsg("could not open file \"%s\" (log file %u, segment %u): %m",
									path, sendId, sendSeg)));
			}
			sendOff = 0;
		}

		/* Need to seek in the file? */
		if (sendOff != startoff)
		{
			if (lseek(sendFile, (off_t) startoff, SEEK_SET) < 0)
				ereport(ERROR,
						(errcode_for_file_access(),
						 errmsg("could not seek in log file %u, segment %u to offset %u: %m",
								sendId, sendSeg, startoff)));
			sendOff = startoff;
		}

		/* How many bytes are within this segment? */
		if (nbytes > (XLogSegSize - startoff))
			segbytes = XLogSegSize - startoff;
		else
			segbytes = nbytes;

		readbytes = read(sendFile, buf, segbytes);
		if (readbytes <= 0)
			ereport(ERROR,
					(errcode_for_file_access(),
			errmsg("could not read from log file %u, segment %u, offset %u, "
				   "length %lu: %m",
				   sendId, sendSeg, sendOff, (unsigned long) segbytes)));

		/* Update state for read */
		XLByteAdvance(recptr, readbytes);

		sendOff += readbytes;
		nbytes -= readbytes;
		buf += readbytes;
	}

	/*
	 * After reading into the buffer, check that what we read was valid. We do
	 * this after reading, because even though the segment was present when we
	 * opened it, it might get recycled or removed while we read it. The
	 * read() succeeds in that case, but the data we tried to read might
	 * already have been overwritten with new WAL records.
	 */
	XLogGetLastRemoved(&lastRemovedLog, &lastRemovedSeg);
	XLByteToSeg(startRecPtr, log, seg);
	if (log < lastRemovedLog ||
		(log == lastRemovedLog && seg <= lastRemovedSeg))
	{
		char		filename[MAXFNAMELEN];

		XLogFileName(filename, ThisTimeLineID, log, seg);
		ereport(ERROR,
				(errcode_for_file_access(),
				 errmsg("requested WAL segment %s has already been removed",
						filename)));
	}
}

/*
 * Read up to MAX_SEND_SIZE bytes of WAL that's been flushed to disk,
 * but not yet sent to the client, and send it.
 *
 * If there is no unsent WAL remaining, *caughtup is set to true, otherwise
 * *caughtup is set to false.
 *
 * If there is pending WAL in output buffer, *pending is set to true,
 * otherwise *pending is set to false.
 *
 * Returns true if OK, false if trouble.
 */
static bool
XLogSend(bool *caughtup, bool *pending)
{
	XLogRecPtr	SendRqstPtr;
	XLogRecPtr	startptr;
	static XLogRecPtr	endptr;
	Size		nbytes;
	uint32		n32;
	int			res;
	WalDataMessageHeader msghdr;

	/* Attempt to flush pending WAL in output buffer */
	if (*pending)
	{
		if (WalSndOutHead != WalSndOutTail)
		{
			res = pq_putbytes_if_writable(WalSndOutBuffer + WalSndOutHead,
										  WalSndOutTail - WalSndOutHead);
			if (res == EOF)
				return false;
			WalSndOutHead += res;
			if (WalSndOutHead != WalSndOutTail)
				return true;
		}

		res = pq_flush_if_writable();
		if (res == EOF)
			return false;
		if (res == 0)
			return true;

		goto updt;
	}

	/*
	 * Attempt to send all data that's already been written out and fsync'd to
	 * disk.  We cannot go further than what's been written out given the
	 * current implementation of XLogRead().  And in any case it's unsafe to
	 * send WAL that is not securely down to disk on the master: if the master
	 * subsequently crashes and restarts, slaves must not have applied any WAL
	 * that gets lost on the master.
	 */
	SendRqstPtr = GetFlushRecPtr();

	/* Quick exit if nothing to do */
	if (XLByteLE(SendRqstPtr, sentPtr))
	{
		*caughtup = true;
		return true;
	}

	/*
	 * Figure out how much to send in one message. If there's no more than
	 * MAX_SEND_SIZE bytes to send, send everything. Otherwise send
	 * MAX_SEND_SIZE bytes, but round back to logfile or page boundary.
	 *
	 * The rounding is not only for performance reasons. Walreceiver relies on
	 * the fact that we never split a WAL record across two messages. Since a
	 * long WAL record is split at page boundary into continuation records,
	 * page boundary is always a safe cut-off point. We also assume that
	 * SendRqstPtr never points to the middle of a WAL record.
	 */
	startptr = sentPtr;
	if (startptr.xrecoff >= XLogFileSize)
	{
		/*
		 * crossing a logid boundary, skip the non-existent last log segment
		 * in previous logical log file.
		 */
		startptr.xlogid += 1;
		startptr.xrecoff = 0;
	}

	endptr = startptr;
	XLByteAdvance(endptr, MAX_SEND_SIZE);
	if (endptr.xlogid != startptr.xlogid)
	{
		/* Don't cross a logfile boundary within one message */
		Assert(endptr.xlogid == startptr.xlogid + 1);
		endptr.xlogid = startptr.xlogid;
		endptr.xrecoff = XLogFileSize;
	}

	/* if we went beyond SendRqstPtr, back off */
	if (XLByteLE(SendRqstPtr, endptr))
	{
		endptr = SendRqstPtr;
		*caughtup = true;
	}
	else
	{
		/* round down to page boundary. */
		endptr.xrecoff -= (endptr.xrecoff % XLOG_BLCKSZ);
		*caughtup = false;
	}

	nbytes = endptr.xrecoff - startptr.xrecoff;
	Assert(nbytes <= MAX_SEND_SIZE);

	/*
	 * OK to read and send the slice.
	 */
	WalSndOutBuffer[0] = 'd';
	WalSndOutBuffer[5] = 'w';
	WalSndOutHead = 0;
	WalSndOutTail = 6 + sizeof(WalDataMessageHeader) + nbytes;

	n32 = htonl((uint32) WalSndOutTail - 1);
	memcpy(WalSndOutBuffer + 1, &n32, 4);

	/*
	 * Read the log directly into the output buffer to avoid extra memcpy
	 * calls.
	 */
	XLogRead(WalSndOutBuffer + 6 + sizeof(WalDataMessageHeader), startptr, nbytes);

	/*
	 * We fill the message header last so that the send timestamp is taken as
	 * late as possible.
	 */
	msghdr.dataStart = startptr;
	msghdr.walEnd = SendRqstPtr;
	msghdr.sendTime = GetCurrentTimestamp();

	memcpy(WalSndOutBuffer + 6, &msghdr, sizeof(WalDataMessageHeader));

	res = pq_putbytes_if_writable(WalSndOutBuffer, WalSndOutTail);
	if (res == EOF)
		return false;

	WalSndOutHead = res;
	if (WalSndOutHead != WalSndOutTail)
	{
		*caughtup = false;
		*pending = true;
		return true;
	}

	/* Flush pending output to the client */
	res = pq_flush_if_writable();
	if (res == EOF)
		return false;
	if (res == 0)
	{
		*caughtup = false;
		*pending = true;
		return true;
	}

updt:
	WalSndOutHead = WalSndOutTail = 0;
	*pending = false;

	sentPtr = endptr;

	/* Update shared memory status */
	{
		/* use volatile pointer to prevent code rearrangement */
		volatile WalSnd *walsnd = MyWalSnd;

		SpinLockAcquire(&walsnd->mutex);
		walsnd->sentPtr = sentPtr;
		SpinLockRelease(&walsnd->mutex);
	}

	/* Report progress of XLOG streaming in PS display */
	if (update_process_title)
	{
		char		activitymsg[50];

		snprintf(activitymsg, sizeof(activitymsg), "streaming %X/%X",
				 sentPtr.xlogid, sentPtr.xrecoff);
		set_ps_display(activitymsg, false);
	}

	return true;
}

/*
 * Attempt to execute standby_fencing_command at the end of replication.
 */
static void
ExecuteStandbyFencingCommand(void)
{
	char		standbyFencingCmd[MAXPGPATH];
	char	   *dp;
	char	   *endp;
	const char *sp;
	int			rc;

	/* Do nothing if no command supplied */
	if (standby_fencing_command[0] == '\0')
		return;

	/*
	 * construct the command to be executed
	 */
	dp = standbyFencingCmd;
	endp = standbyFencingCmd + MAXPGPATH - 1;
	*endp = '\0';

	for (sp = standby_fencing_command; *sp; sp++)
	{
		if (*sp == '%')
		{
			switch (sp[1])
			{
				case 'a':
				{
					/* %a: application_name */
					const char *appname = application_name;

					if (appname == NULL || *appname == '\0')
						appname = _("[unknown]");

					sp++;
					strlcpy(dp, appname, endp - dp);
					dp += strlen(dp);
					break;
				}
				case '%':
					/* convert %% to a single % */
					sp++;
					if (dp < endp)
						*dp++ = *sp;
					break;
				default:
					/* otherwise treat the % as not special */
					if (dp < endp)
						*dp++ = *sp;
					break;
			}
		}
		else
		{
			if (dp < endp)
				*dp++ = *sp;
		}
	}
	*dp = '\0';

	ereport(DEBUG3,
			(errmsg_internal("executing standby fencing command \"%s\"",
							 standbyFencingCmd)));

	/*
	 * execute the constructed command
	 */
	rc = system(standbyFencingCmd);
	if (rc != 0)
	{
		/*
		 * No matter what code is returned, walsender can't stop exiting.
		 * We don't need to care about the return code of the command here.
		 */
		ereport(WARNING,
				(errmsg("standby fencing command failed with return code %d",
						rc),
				 errdetail("The failed standby fencing command was: %s",
						   standbyFencingCmd)));
	}
}

/* SIGHUP: set flag to re-read config file at next convenient time */
static void
WalSndSigHupHandler(SIGNAL_ARGS)
{
	got_SIGHUP = true;
	if (MyWalSnd)
		SetLatch(&MyWalSnd->latch);
}

/*
 * SIGINT: set flag to prevent the server from getting into standalone mode.
 *
 * Since the flag is used only when the walsender exits, we don't need to
 * wake it up soon.
 */
static void
WalSndDisableStandaloneHandler(SIGNAL_ARGS)
{
	standalone_disabled = true;
}

/* SIGTERM: set flag to shut down */
static void
WalSndShutdownHandler(SIGNAL_ARGS)
{
	shutdown_requested = true;
	if (MyWalSnd)
		SetLatch(&MyWalSnd->latch);
}

/*
 * WalSndQuickDieHandler() occurs when signalled SIGQUIT by the postmaster.
 *
 * Some backend has bought the farm,
 * so we need to stop what we're doing and exit.
 */
static void
WalSndQuickDieHandler(SIGNAL_ARGS)
{
	PG_SETMASK(&BlockSig);

	/*
	 * We DO NOT want to run proc_exit() callbacks -- we're here because
	 * shared memory may be corrupted, so we don't want to try to clean up our
	 * transaction.  Just nail the windows shut and get out of town.  Now that
	 * there's an atexit callback to prevent third-party code from breaking
	 * things by calling exit() directly, we have to reset the callbacks
	 * explicitly to make this work as intended.
	 */
	on_exit_reset();

	/*
	 * Note we do exit(2) not exit(0).	This is to force the postmaster into a
	 * system reset cycle if some idiot DBA sends a manual SIGQUIT to a random
	 * backend.  This is necessary precisely because we don't clean up our
	 * shared memory state.  (The "dead man switch" mechanism in pmsignal.c
	 * should ensure the postmaster sees this as a crash, too, but no harm in
	 * being doubly sure.)
	 */
	exit(2);
}

/* SIGUSR1: set flag to send WAL records */
static void
WalSndXLogSendHandler(SIGNAL_ARGS)
{
	latch_sigusr1_handler();
}

/* SIGUSR2: set flag to do a last cycle and shut down afterwards */
static void
WalSndLastCycleHandler(SIGNAL_ARGS)
{
	ready_to_stop = true;
	if (MyWalSnd)
		SetLatch(&MyWalSnd->latch);
}

/* Set up signal handlers */
void
WalSndSignals(void)
{
	/* Set up signal handlers */
	pqsignal(SIGHUP, WalSndSigHupHandler);		/* set flag to read config
												 * file */
	pqsignal(SIGINT, WalSndDisableStandaloneHandler);	/* disable standalone */
	pqsignal(SIGTERM, WalSndShutdownHandler);	/* request shutdown */
	pqsignal(SIGQUIT, WalSndQuickDieHandler);	/* hard crash time */
	pqsignal(SIGALRM, SIG_IGN);
	pqsignal(SIGPIPE, SIG_IGN);
	pqsignal(SIGUSR1, WalSndXLogSendHandler);	/* request WAL sending */
	pqsignal(SIGUSR2, WalSndLastCycleHandler);	/* request a last cycle and
												 * shutdown */

	/* Reset some signals that are accepted by postmaster but not here */
	pqsignal(SIGCHLD, SIG_DFL);
	pqsignal(SIGTTIN, SIG_DFL);
	pqsignal(SIGTTOU, SIG_DFL);
	pqsignal(SIGCONT, SIG_DFL);
	pqsignal(SIGWINCH, SIG_DFL);
}

/* Report shared-memory space needed by WalSndShmemInit */
Size
WalSndShmemSize(void)
{
	Size		size = 0;

	size = offsetof(WalSndCtlData, walsnds);
	size = add_size(size, mul_size(max_wal_senders, sizeof(WalSnd)));

	/*
	 * If replication is enabled, we have a data structure called
	 * WalSndWaiters, created in shared memory.
	 */
	if (max_wal_senders > 0)
		size = add_size(size, mul_size(MaxBackends, sizeof(WalSndWaiter)));

	return size;
}

/* Allocate and initialize walsender-related shared memory */
void
WalSndShmemInit(void)
{
	bool		found;
	int			i;
	Size		size = add_size(offsetof(WalSndCtlData, walsnds),
								mul_size(max_wal_senders, sizeof(WalSnd)));

	WalSndCtl = (WalSndCtlData *)
		ShmemInitStruct("Wal Sender Ctl", size, &found);

	if (!found)
	{
		/* First time through, so initialize */
		MemSet(WalSndCtl, 0, size);

		for (i = 0; i < max_wal_senders; i++)
		{
			WalSnd	   *walsnd = &WalSndCtl->walsnds[i];

			SpinLockInit(&walsnd->mutex);
			InitSharedLatch(&walsnd->latch);
		}
	}

	/* Create or attach to the WalSndWaiters array too, if needed */
	if (max_wal_senders > 0)
	{
		WalSndWaiters = (WalSndWaiter *)
			ShmemInitStruct("WalSndWaiters",
							mul_size(MaxBackends, sizeof(WalSndWaiter)),
							&found);
		WalSndCtl->maxWaiters = MaxBackends;
	}
}

/* Wake up all walsenders */
void
WalSndWakeup(void)
{
	int		i;

	for (i = 0; i < max_wal_senders; i++)
		SetLatch(&WalSndCtl->walsnds[i].latch);
}

/*
 * Ensure that replication has been completed up to the given position.
 */
void
WaitXLogSend(XLogRecPtr record)
{
	int		i;
	bool	mustwait = false;

	Assert(max_wal_senders > 0);

	for (i = 0; i < max_wal_senders; i++)
	{
		/* use volatile pointer to prevent code rearrangement */
		volatile WalSnd *walsnd = &WalSndCtl->walsnds[i];
		XLogRecPtr		recptr;
		WalSndState		state;

		if (walsnd->pid == 0)
			continue;

		SpinLockAcquire(&walsnd->mutex);
		state = walsnd->walSndState;
		recptr = walsnd->ackdPtr;
		SpinLockRelease(&walsnd->mutex);

		if (state <= WALSND_ASYNC ||
			(recptr.xlogid == 0 && recptr.xrecoff == 0))
			continue;

		/* Quick exit if already known replicated */
		if (XLByteLE(record, recptr))
			return;

		/*
		 * If walsender is bulk-sending WAL for standby to catch up,
		 * we don't need to wait for Ack from standby.
		 */
		if (state <= WALSND_CATCHUP)
			continue;

		mustwait = true;
	}

	/*
	 * Don't need to wait for replication if there is no synchronous
	 * standby
	 */
	if (!mustwait)
		return;

	/*
	 * Register myself into the wait list and sleep until replication
	 * has been completed up to the given position and the walsender
	 * signals me.
	 *
	 * If replication has been completed up to the latest position
	 * before the registration, walsender might be unable to send the
	 * signal immediately. We must wake up the walsender after the
	 * registration.
	 */
	ResetLatch(&MyProc->latch);
	RegisterWalSndWaiter(MyBackendId, record, &MyProc->latch);
	WalSndWakeup();

	for (;;)
	{
		WaitLatch(&MyProc->latch, 1000000L);

		/* If done already, we finish waiting */
		if (replication_complete)
		{
			replication_complete = false;
			return;
		}

		/*
		 * If the primary cannot work any longer, we don't return
		 * a success to the client
		 */
		if (exit_before_complete)
			return;
	}
}

/*
 * Register the given backend into the wait list.
 */
static void
RegisterWalSndWaiter(BackendId backendId, XLogRecPtr record, Latch *latch)
{
	/* use volatile pointer to prevent code rearrangement */
	volatile WalSndCtlData	*walsndctl = WalSndCtl;
	int		i;
	int		count = 0;

	LWLockAcquire(WalSndWaiterLock, LW_EXCLUSIVE);

	/* Out of slots. This should not happen. */
	if (walsndctl->numWaiters + 1 > walsndctl->maxWaiters)
		elog(PANIC, "out of replication waiters slots");

	/*
	 * The given position is expected to be relatively new in the
	 * wait list. Since the entries in the list are sorted in an
	 * increasing order of XLogRecPtr, we can shorten the time it
	 * takes to find an insert slot by scanning the list backwards.
	 */
	for (i = walsndctl->numWaiters; i > 0; i--)
	{
		if (XLByteLE(WalSndWaiters[i - 1].record, record))
			break;
		count++;
 	}

	/* Shuffle the list if needed */
	if (count > 0)
		memmove(&WalSndWaiters[i + 1], &WalSndWaiters[i],
				count * sizeof(WalSndWaiter));

	WalSndWaiters[i].backendId = backendId;
	WalSndWaiters[i].record = record;
	WalSndWaiters[i].latch = latch;
	walsndctl->numWaiters++;

	LWLockRelease(WalSndWaiterLock);
}

/*
 * Wake up the backends waiting until replication has been completed
 * up to the position older than or equal to the given one.
 *
 * If abort is true, we signal the waiters to exit with FATAL error
 * before returning a success to the client.
 *
 * Wake up all waiters if InvalidXLogRecPtr is given.
   */
static void
WakeupWalSndWaiters(XLogRecPtr record, bool abort)
{
	/* use volatile pointer to prevent code rearrangement */
	volatile WalSndCtlData	*walsndctl = WalSndCtl;
	int		i;
	int		count = 0;
	bool	all_wakeup = (record.xlogid == 0 && record.xrecoff == 0);

	LWLockAcquire(WalSndWaiterLock, LW_EXCLUSIVE);

	for (i = 0; i < walsndctl->numWaiters; i++)
	{
		/* use volatile pointer to prevent code rearrangement */
		volatile WalSndWaiter  *waiter = &WalSndWaiters[i];

		if (all_wakeup || XLByteLE(waiter->record, record))
		{
			SetProcLatch(waiter->latch, abort ? PROCSIG_EXIT_BEFORE_COMPLETE :
						 PROCSIG_REPLICATION_COMPLETE, waiter->backendId);
			count++;
		}
		else
		{
			/*
			 * If the backend waiting for the Ack position newer than
			 * the given one is found, we don't need to search the wait
			 * list any more. This is because the waiters in the list
			 * are guaranteed to be sorted in an increasing order of
			 * XLogRecPtr.
			 */
			break;
		}
	}

	/* If there are still some waiters, left-justify them in the list */
	walsndctl->numWaiters -= count;
	if (walsndctl->numWaiters > 0 && count > 0)
		memmove(&WalSndWaiters[0], &WalSndWaiters[i],
				walsndctl->numWaiters * sizeof(WalSndWaiter));

	LWLockRelease(WalSndWaiterLock);
}

/*
 * Returns the oldest Ack position in synchronous walsenders. Or
 * InvalidXLogRecPtr if none.
 */
static XLogRecPtr
GetOldestAckdPtr(void)
{
	XLogRecPtr	oldest = {0, 0};
	int		i;
	bool	found = false;

	for (i = 0; i < max_wal_senders; i++)
	{
		/* use volatile pointer to prevent code rearrangement */
		volatile WalSnd *walsnd = &WalSndCtl->walsnds[i];
		XLogRecPtr		recptr;
		WalSndState		state;

		if (walsnd->pid == 0)
			continue;

		SpinLockAcquire(&walsnd->mutex);
		state = walsnd->walSndState;
		recptr = walsnd->ackdPtr;
		SpinLockRelease(&walsnd->mutex);

		if (state <= WALSND_ASYNC ||
			(recptr.xlogid == 0 && recptr.xrecoff == 0))
			continue;

		if (!found || XLByteLT(recptr, oldest))
			oldest = recptr;
		found = true;
	}
	return oldest;
}

/*
 * This is called when PROCSIG_REPLICATION_COMPLETE is received.
 */
void
HandleReplicationComplete(void)
{
	replication_complete = true;
}

/*
 * This is called when PROCSIG_EXIT_BEFORE_COMPLETE is received.
 */
void
HandleExitBeforeComplete(void)
{
	exit_before_complete = true;
}

/*
 * Does the backend have to exit before returning a CommandComplete
 * message to the client?
 */
bool
NeedExitBeforeComplete(void)
{
	return exit_before_complete;
}
